#ifndef CORRESPONDENCE_H
#define CORRESPONDENCE_H

#include "Preprocessing.h"

struct assignment {
    MyMesh::VertexHandle s;
    MyMesh::VertexHandle t;
};

namespace Correspondence {

/*
 * Fonction principale de calcul des correspondences entre la source et le model
 * Renvoie un pointeur sur le tableau de correspondence.
 * Effectue les etapes:
 *
 * 1- Initial correspondences:
 *         a) Closest Points
 *         b) Improve correspondence
 * 2- Correspondence Prunning
 * 3- Correspondence Propagation
 * 4- Calcul des poids wi associes a chaque correspondence i
 *
 * Si (*correspondence)[0][i] est un indice du sommet de la source, son correspondant se trouve
 * a  (*correspondence)[1][i] pour tout entier i entre 0 et nb_source-1 (avec nb_source nombre
 * d'element du maillage source)
 * Le tableau est alloue a l'interieur de la fonction
 *
 */
void Correspondence_Computation(MyMesh *source, MyMesh *model, int nb_source, int nb_model, MyMesh::VertexHandle *** correspondence, double c0, double** poids);


/* Correspondence initiales */
void Initial_Correspondence(MyMesh * source, MyMesh * model, int nb_source_subset, MyMesh::VertexHandle *** correspondence);

/* Correspondence Prunning:
 *
 * 1- Calcul de la matrice de consistence
 * 2- standard spectral matching Method (LH05)
 *
 */
void Prunning(int nb_vertices, MyMesh::VertexHandle *** correspondence, double c0, std::list<assignment> *kernel, int **has_a_correspondence);

/*
 * Calcul de la matrice de consistence
 */
void Consistent_Matrix(int nb_vertices, MyMesh::VertexHandle ** correspondence, gsl_matrix **m, double c0);

/*
 * Measure de consistence utilisee dans l'algorithme de Correspondence Prunning
 *
 */
double Consistency_Measure(MyMesh::VertexHandle si, MyMesh::VertexHandle ti, MyMesh::VertexHandle sj, MyMesh::VertexHandle tj);

/*
 * Algorithme de propagation a partir d'un noyau de correspondence 'kernel' :
 * Pour chaque sommets de la source n'ayant pas de correspondant ( valeur a 0 dans le tableau has_a_correspondence
 * qui doit etre fourni 1 sinon), on assigne un sommets dans la cible avec la fonction Assign
 *
 */
void Propagation(int nb_vertices, std::list<assignment> *kernel, int **has_a_correspondence, MyMesh::VertexHandle *** correspondence);

/*
 * Assigne un sommet cible ti  a un sommet si n'ayant pas de correspondant
 * La fonction renvoit l'index (VertexHandle) du sommet ti
 * Cette fonction calcule l'erreur de consistence entre deux sommets s et t avec la fonction
 * Consistency_Error
 */
MyMesh::VertexHandle Assign(std::list<assignment> *kernel, int si);

/*
 * Calcul de l'erreur de consistence entre deux element s et t
 *
 */
double Consistency_Error(std::list<assignment> *kernel, int s, int t);


void Fine_Tuning(MyMesh *source , MyMesh  * model, MyMesh::VertexHandle ** correspondence);

/*
 * Amelioration des correspondances.
 * Cette fonction ameliore les correspondences (tableau correspodence) a l'aide des ensemble de voisins
 * (neighbors) calcule dans la partie de Preprocessing
 *
 */
void Improve_Correspondence(MyMesh *source, MyMesh *model, int nb_source, MyMesh::VertexHandle ** neighbors, double threshold, MyMesh::VertexHandle *** correspondence, int etape);

/*
 * Norme euclidienne entre deux points d'indices p1 et p2 de maillage respectif *source et *model
 *
 */
double Normal(MyMesh *source, MyMesh *model , MyMesh::VHandle p1, MyMesh::VHandle p2);

/*
 * Norme euclidienne entre deux vecteur v1 et v2 de taille n
 *
 */
double Norm(double *v1, double *v2, int n);

/*
 * Norme euclidienne entre deux vecteurs de points p1[] et p2[] de taille n
 *
 */
double Norm(MyMesh::Point *v1, MyMesh::Point *v2, int n);

/*
 * Fonction permettant l'acces au point du maillage *m correspondant a l'indice num_vertice
 *
 */
MyMesh::Point Access_Point(MyMesh *m, MyMesh::VertexHandle num_vertice);


/*
 * Calcul des poids pour chaque correspondence a partir du noyau de correspondences calculees
 * dans la partie Correspondence Prunning
 *
 */
void Poids(MyMesh::VertexHandle ** correspondence, std::list<assignment> *kernel, int nb_vertices, double **resultat);


void Affect_Preprocessing_Correspondence(double **distance_1,  double **distance_2,  MyMesh::VHandle **voisins_1,  MyMesh::VHandle **voisins_2, double diag, double **tab_feature_vector_source, double **tab_feature_vector_model);

}
#endif
