#ifndef ICP2_H
#define ICP2_H


#include <iostream>
#include <fstream>
#include <limits>
#include <cmath>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_matrix.h>

// -------------------- OpenMesh
#include <OpenMesh/Core/IO/MeshIO.hh>
#include <OpenMesh/Core/Mesh/Types/TriMesh_ArrayKernelT.hh>
#include <OpenMesh/Core/Mesh/Types/PolyMesh_ArrayKernelT.hh>
#include <list>

typedef OpenMesh::PolyMesh_ArrayKernelT<>  MyPolyMesh;
typedef OpenMesh::TriMesh_ArrayKernelT<> MyMesh;


namespace icp2 {

/* Fonction principale: Corresponding Point Set Registration (A Method for Registration of 3-D Shapes" [BM92])
 *
 * Calcule la transformation representee par la matrice 4*4 transformation_out a partir du sous ensemble de
 * sommets du maillage source et du sous ensemble de leurs correspondants du maillage model.
 * Les deux sous ensembles sont representes par leur indice (VertexHandle) dans leur maillage correspondant.
 * et doivent avoir le meme nombre de points (nb_vertices)
 *
 * Chaque indice source_handle[i] doit avoir son correspondant dans model_handle[i]
 * pour tout i dans {0,1,..,nv_vertices-1}
 *
 * La tableau transformation_out est alloue a l'interieur de la fonction
 *
 */
void Align (MyMesh *source, MyMesh::VertexHandle *source_handle, MyMesh *model, MyMesh::VertexHandle *model_handle, double*** transformation_out, double *erreur, int nb_vertices);

/*
 * Calcul du centre des masses &u a partir de nb_vertices indices de points x appartenant au maillage &m .
 * Le vecteur renvoye representant le centre des masses est alloue a l'interieur de la fonction.
 *
 */
void Center_Mass(MyMesh *mesh, MyMesh::VertexHandle *x, gsl_vector ** u, int nb_vertices);

/*
 * Calcul de la matrice de covariance &sigmae_px a partir d'un ensemble de nb_vertices indices de sommets
 * p_handle du maillage &p et d'un ensemble de nb_vertices indices de sommets x_handle du maillage &x
 * et des centre de masse up et ux correspondant.
 * La matrice renvoyee est alloue a l'interieur de la fonction.
 *
 */
void Cross_Covariance(MyMesh *p, MyMesh::VertexHandle *p_handle, MyMesh *x, MyMesh::VertexHandle *x_handle, gsl_vector * up, gsl_vector * ux, gsl_matrix ** sigma_px, int nb_vertices);


/*
 * Calcul du vecteur &delta a partir de la matrice m
 * Le vecteur renvoyee est alloue a l'interieur de la fonction
 *
 */
void Delta_M(gsl_matrix * m, gsl_vector ** delta);

/*
 * Calcul de la matrice anti-symetrique &q_matrix a partir de la matrice de covariance sigma_x et du vecteur delta.
 * La matrice renvoyee est alloue a l'interieur de la fonction
 *
 */
void Q_Matrix(gsl_matrix * sigma_px, gsl_vector * delta, gsl_matrix ** q_matrix);

/*
 * Calcul du vecteur propre & v associe a la plus grande valeur propre de la matrice m
 * Le vecteur renvoye est alloue a l'interieur de la fonction
 *
 */
void Eigen_V(gsl_matrix * m, gsl_vector ** v, int n);

/*
 * Fonction calculant la matrice de rotation *Rqr associee au quaternion represente par le vecteur
 * de dimension 4 qr.
 * La matrice renvoye est alloue a l'interieur de la fonction
 *
 */
void Q_Rotation(gsl_vector * qr, gsl_matrix ** Rqr);

/*
 * Copy les coordonnees du point p de dimension n dans le vecteur u
 *
 */
void Mem_Copy_Point_V(MyMesh::Point p, gsl_vector * u, int n);

/*
 * Renvoit le point associe a l'indice num_vertice du maillage &m
 *
 */
MyMesh::Point Access_Point(MyMesh *m, MyMesh::VertexHandle num_vertice);

/*
 * Effectue l'operation u * transpose(v) ou u et v representent des vecteurs colonne de dimension 3
 * Retourne une matrice *m 3*3 alloue a l'interieur de la fonction.
 */
void Mult_Matrice(gsl_vector * u, gsl_vector * v, gsl_matrix ** m);

/*
 * Effectue l'operation m*v ou m est une matrice de dimension n*n  et v un vecteur de dimension n
 * Le resultat est stocke dans v;
 *
 */
void Mult_Matrice_Vector(gsl_matrix* m, gsl_vector * v, int n);

/*
 *
 * Affiche une matrice de dimension n1*n2
 *
 */
void Display_Matrix(gsl_matrix * m, int n1, int n2);

/*
 *
 * Affiche une vecteur de dimension n
 *
 */
void Display_Vector(gsl_vector *v, int n);

/*
*
* Affiche la transformation (matrice 4*4)
*
*/
void Display_Transf (double** transformation);

/*
 * Soit yi le point d'indice i du sous ensemble d'indices source_handle du maillage &source et
 * xi_t le point obtenu apres transformation appliquee au point xi d'indice i du sous ensemble
 * d'indices model_handle du maillage &model
 *
 * La fonction calcul et renvoie la valeur egale a la somme des norme euclidiennes des differences
 * entre chaque point yi et xi_t divisee par le nombre de points nb_vertices
 *
 */
double mean_square_error(MyMesh *source, MyMesh::VertexHandle *source_handle, MyMesh *model, MyMesh::VertexHandle *model_handle, double** transformation, int nb_vertices);

/*
 * Recopie les nb_vertices indices des points du maillage &m dans le tableau d'indices &m_handle
 * Le tableau renvoye est alloue a l'interieur de la fonction
 *
 */
void Copie_Mesh_Handle(MyMesh *m, MyMesh::VertexHandle ** m_handle, int nb_vertices);

void Closest_Points (MyMesh *source, MyMesh *model, int nb_source, MyMesh::VertexHandle *** resultat);
}

#endif
